//* swahili.cpp
//*********************************************************************
//*
//*  SWAHILI - Kiswahili to English translation program
//*  Copyright (C) 1997-1998 Morris David Fried
//*
//*  This program is free software; you can redistribute it and/or
//*  modify it under the terms of the GNU General Public License
//*  as published by the Free Software Foundation; either version 2
//*  of the License, or (at your option) any later version.
//*
//*  This program is distributed in the hope that it will be useful,
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//*  GNU General Public License for more details.
//*
//*  You should have received a copy of the GNU General Public License
//*  along with this program; if not, write to the Free Software
//*  Foundation, Inc., 59 Temple Place - Suite 330,
//*  Boston, MA 02111-1307, USA.
//*
//*  I welcome any questions or comments about this software and I can
//*  be contacted at HAWKEYEOP@JUNO.COM or at 23 Lincoln Road North,
//*  Plainview, NY 11803, USA.
//*
//*********************************************************************
//* file includes
#include <fstream.h>
#include <stdio.h>
#include <iostream.h>
#include <string.h>
//* Preprocessor variables
#define CONTROL_DATA_SIZE 32
#define NOUNS_ENTRY_SIZE 48
#define VERBS_ENTRY_SIZE 120
#define ADJETC_ENTRY_SIZE 44
#define IRRPLU_ENTRY_SIZE 20
//* function prototypes
char plurals (void);    /* look for noun plurals */
char adjform (void);    /* look for adjective forms */
char specform (void);    /* look for special word forms */
char thisthat (void);    /* look for forms of this and that */
char verbform (void);    /* look for verb forms */
int trsnoun (char word_in[]);    /* translate noun */
int trsverb (char word_in[]);    /* translate verb */
int trsadj (char word_in[]);     /* translate adjective or other */
int irrplu (int index_in);    /* read in English irregular plural */
//* dictionary structures
  struct noun_entry
  {
    char swa_word[19];
    char eng_word[19];
    int short tie_breaker;
    int short plural_index;
    char grammar_flag;
    char plural_prefix[5];
  };
  struct noun_entry dict_noun;
  struct verb_entry
  {
    char swa_word[19];
    char eng_word[19];
    int short tie_breaker;
    int short filler1;
    char grammar_flag;
    char eng_ing_form[19];
    char eng_s_form[19];
    char eng_past[19];
    char eng_participle[19];
    char filler2;
  };
  struct verb_entry dict_verb;
  struct adjetc_entry
  {
    char swa_word[19];
    char eng_word[19];
    int short tie_breaker;
    int short forms_type;
    char speech_part;
    char grammar_flag;
  };
  struct adjetc_entry dict_adjetc;
  char part_speech[20];    /* to hold meaning of speech_part */
  struct control_record
  {
    char program_version[4];
    char version_date[8];
    int tot_swahili_words;
    int noun_dict_entries;
    int verb_dict_entries;
    int adjetc_dict_entries;
    int eng_irreg_plu_entries;
  };
  struct control_record control_data;
  char eng_irreg_plural [20];
//* File variables
  FILE * ctl_file;      /*  control file  */
  FILE * noun_dict;     /*  nouns dictionary  */
  FILE * verb_dict;     /*  verbs dictionary  */
  FILE * adjetc_dict;   /*  adjectives etc. dictionary  */
  FILE * irrplu_dict;   /*  English irregular plurals  */
//* other common variables
  char swahili_word[16];
  char swahili_hold[16];
  char english_word[18];    /* leave extra room for forms */
  char english_hold[18];
  int dict_pos, eng_index;
  int i;
//* Main Routine
main()
{
//* variables
  char found_flag;
  char of_subject[21];
//* executable code
//* display copyright information
  cout << "\n" << "SWAHILI - Kiswahili to English translation program.";
  ctl_file = fopen ("SWAHILI.CTL","rb");
  if (ctl_file == NULL)
  {
    cout << "\n" << "SWAHILI.CTL does not exist: run NOUNS first";
    cout << "\n" << "then run VERBS and ADJETC to complete file."
      << "\n";
    return 12;
  }
  fread (&control_data, CONTROL_DATA_SIZE, 1, ctl_file);
  fclose (ctl_file);
  cout << "\n" << "Version " << control_data.program_version
    << " as of " << control_data.version_date
    << " with a vocabulary of "
    << control_data.tot_swahili_words << " words.";
  cout << "\n" << "COPYRIGHT (C) 1997-1998 MORRIS DAVID FRIED";
  cout << "\n" << "SWAHILI comes with ABSOLUTELY NO WARRANTY; "
    << "for details type 'qw'.";
  cout << "\n" << "This is free software, and you are welcome to "
    << "redistribute it ";
  cout << "\n" << "under certain conditions; type 'qc' "
    << "for details." << "\n";
  noun_dict = fopen ("NOUNS.DCY","rb");
  verb_dict = fopen ("VERBS.DCY","rb");
  adjetc_dict = fopen ("ADJETC.DCY","rb");
  irrplu_dict = fopen ("IRRPLU.DCY","rb");
//* ASK-FOR-A-WORD loop
ask:
  cout << "\n" << "Please enter a Kiswahili word (or 'quit') ... ";
  cin >> swahili_word;
//*  Convert all input to lower-case */
  for (i = 0; i < 16; ++i)
    if (swahili_word[i] > '\x40' && swahili_word[i] < '\x5B')
      swahili_word[i] = swahili_word[i] + '\x20';
  if (swahili_word[0] == 'q')    /* no q in Swahili */
  {
    if (swahili_word[1] == 'w')    /* qw = display warranty */
    {
      cout << "\n" << "BECAUSE THIS PROGRAM IS LICENSED FREE OF "
        << "CHARGE, THERE IS NO WARRANTY";
      cout << "\n" << "FOR THE PROGRAM TO THE EXTENT PERMITTED BY "
        << "APPLICABLE LAW.  EXCEPT WHEN";
      cout << "\n" << "OTHERWISE STATED IN WRITING THE COPYRIGHT "
        << "HOLDERS AND/OR OTHER PARTIES";
      cout << "\n" << "PROVIDE THE PROGRAM 'AS IS' WITHOUT WARRANTY "
        << "OF ANY KIND, EITHER EXPRESSED";
      cout << "\n" << "OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, "
        << "THE IMPLIED WARRANTIES OF";
      cout << "\n" << "MERCHANTABILITY AND FITNESS FOR A PARTICULAR "
        << "PURPOSE.  THE ENTIRE RISK AS";
      cout << "\n" << "TO THE QUALITY AND PERFORMANCE OF THE PROGRAM "
        << "IS WITH YOU.  SHOULD THE";
      cout << "\n" << "PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST "
        << "OF ALL NECESSARY SERVICING,";
      cout << "\n" << "REPAIR OR CORRECTION." << "\n";
    }
    else if (swahili_word[1] == 'c')    /* qc = display copy rules */
    {
      cout << "\n" << "You may copy and distribute verbatim copies "
        << "of this program and its";
      cout << "\n" << "source code as you receive it, in any medium, "
        << "provided that you";
      cout << "\n" << "conspicuously and appropriately publish on "
        << "each copy an appropriate";
      cout << "\n" << "copyright notice and disclaimer of warranty; "
        << "keep intact all the";
      cout << "\n" << "notices that refer to the GNU GENERAL PUBLIC "
        << "LICENSE and to the absence";
      cout << "\n" << "of any warranty; and give any other "
        << "recipients of this program";
      cout << "\n" << "a copy of the GNU GENERAL PUBLIC LICENSE along "
        << "with this program." << "\n";
    }
    else
    {
      fclose (noun_dict);
      fclose (verb_dict);
      fclose (adjetc_dict);
      fclose (irrplu_dict);
      return 0;
    }
    cout << "\n" << "See the GNU GENERAL PUBLIC LICENSE for complete "
      << "details.";
    cout << "\n" << "You should have received a copy of the GNU "
      << "General Public License along";
    cout << "\n" << "with this program; if not, write to the Free "
      << "Software Foundation, Inc.,";
    cout << "\n" << "59 Temple Place - Suite 330, Boston, MA "
      << "02111-1307, USA." << "\n";
    goto ask;
  }
  found_flag = 'n';
//* Is this word a singular noun?
  dict_pos = trsnoun(swahili_word);
  if (dict_pos > -1)
  {
    cout << "\n" << dict_noun.eng_word;
    cout << "\n" << "noun" << "\n";
    found_flag = 'y';
  }
//*  Is this word the root of an verb?
  dict_pos = trsverb(swahili_word);
  if (dict_pos > -1)
  {
    cout << "\n" << dict_verb.eng_word << ", ";
//*  Make first letter of English upper-case, for imperative
    if (dict_verb.eng_word[0] > '\x60' &&
      dict_verb.eng_word[0] < '\x7B')
        dict_verb.eng_word[0] = dict_verb.eng_word[0] - '\x20';
    cout << dict_verb.eng_word << "!";
    cout << "\n" << "verb - root, imperative" << "\n";
    found_flag = 'y';
  }
//*  Is this word an adjective root or another part of speech?
  dict_pos = trsadj(swahili_word);
  if (dict_pos > -1)
  {
    cout << "\n" << dict_adjetc.eng_word << "\n";
    if (part_speech[0] != 'x')  cout << part_speech << "\n";
    found_flag = 'y';
  }
//*  Save unaltered word.
  for (i = 0; i < 16; ++i)
    swahili_hold[i] = swahili_word[i];
//*  Check for forms of words - first, plurals of nouns.
  if (plurals() == 'y')  found_flag = 'y';
//*  Check for adjectives modifying nouns.
  if (adjform() == 'y')  found_flag = 'y';
//*  Check for special words modifying nouns.
  if (specform() == 'y')  found_flag = 'y';
//*  Check for forms of this and that.
  if (thisthat() == 'y')  found_flag = 'y';
//*  Check for forms of verbs.
  if (verbform() == 'y')  found_flag = 'y';
  if (found_flag == 'n')
    cout << "\n" << "Sorry, that word is not in my dictionary" << "\n";
  goto ask;
  return 0;
}
//*  Noun plural check routine
char plurals (void)
{
  char plural_flag = 'n';
  if (swahili_hold[0] == 'v' && swahili_hold[1] == 'i')
  {
    swahili_word[0] = 'k';    /*  KI/VI  */
    dict_pos = trsnoun(swahili_word);
    if (dict_pos > -1)  plural_flag = 'y';
  }
  else if (swahili_hold[0] == 'v' && swahili_hold[1] == 'y')
  {
    swahili_word[0] = 'c';    /*  CH/VY  */
    swahili_word[1] = 'h';
    dict_pos = trsnoun(swahili_word);
    if (dict_pos > -1)  plural_flag = 'y';
  }
  else if (swahili_hold[0] == 'm' && swahili_hold[1] == 'i')
  {
    if (swahili_hold[2] == 'a' || swahili_hold[2] == 'e'
      || swahili_hold[2] == 'i')
      swahili_word[1] = 'w';    /*  MW/MI  */
    else if (swahili_hold[2] == 'h' || swahili_hold[2] == 'w')
      swahili_word[1] = 'u';    /*  MU/MI  */
    else                        /*  M/MI  */
    {
      for (i = 1; i < 15; ++i)
        swahili_word[i] = swahili_hold[i+1];   /*remove 'i'*/
      swahili_word[15] = '\x00';   /* mark end of word */
    }
    dict_pos = trsnoun(swahili_word);
    if (dict_pos > -1)  plural_flag = 'y';
  }
  else if (swahili_hold[0] == 'w' && swahili_hold[1] == 'a')
  {
    if ((swahili_hold[2] == 'i') || (swahili_hold[2] == 'o'))
    {
      swahili_word[0] = 'm';    /*  MW/WA  */
      swahili_word[1] = 'w';
      dict_pos = trsnoun(swahili_word);
      if (dict_pos > -1)  plural_flag = 'y';
    }
    else
    {
      swahili_word[0] = 'm';    /*  M/WA  */
      for (i = 1; i < 15; ++i)
        swahili_word[i] = swahili_hold[i+1];   /*remove 'a'*/
      swahili_word[15] = '\x00';   /* mark end of word */
      dict_pos = trsnoun(swahili_word);
      if (dict_pos > -1)  plural_flag = 'y';
      else
      {
        swahili_word[0] = 'm';    /*  MWA/WA  */
        for (i = 1; i < 16; ++i)
          swahili_word[i] = swahili_hold[i-1];   /*'m' plus original*/
        dict_pos = trsnoun(swahili_word);
        if (dict_pos > -1)  plural_flag = 'y';
      }
    }
  }
  else if (swahili_hold[0] == 'w' && swahili_hold[1] == 'e')
  {
    if (swahili_hold[2] == 'n')
    {
      swahili_word[0] = 'm';    /*  MWE/WE  */
      for (i = 1; i < 16; ++i)
        swahili_word[i] = swahili_hold[i-1];   /*'m' plus original*/
      dict_pos = trsnoun(swahili_word);
      if (dict_pos > -1)  plural_flag = 'y';
    }
    else if (swahili_hold[2] == 'v' || swahili_hold[2] == 'z')
    {
      swahili_word[0] = 'm';    /*  MWI/WE  */
      for (i = 1; i < 16; ++i)
        swahili_word[i] = swahili_hold[i-1];   /*'m' plus original*/
      swahili_word[2] = 'i';    /*  change third letter  */
      dict_pos = trsnoun(swahili_word);
      if (dict_pos > -1)  plural_flag = 'y';
    }
  }
  else if ((swahili_hold[0] == 'm') &&    /*  MA  */
           (swahili_hold[1] == 'a' || swahili_hold[1] == 'e'))
  {
    swahili_word[0] = 'j';    /*  JI/MA - some MA-/ME- plurals  */
    swahili_word[1] = 'i';    /*    have JI- singulars    */
    dict_pos = trsnoun(swahili_word);
    if (dict_pos > -1)  plural_flag = 'y';
    else if (swahili_hold[1] == 'a')
    {                         /*  or even JA/MA  */
      swahili_word[1] = 'a';
      dict_pos = trsnoun(swahili_word);
      if (dict_pos > -1)  plural_flag = 'y';
      else
      {       /* most MA- plurals just drop the MA */
        for (i = 0; i < 14; ++i)
          swahili_word[i] = swahili_hold[i+2];   /*word w/o ma-*/
        swahili_word[14] = '\x00';   /* mark end of word */
        dict_pos = trsnoun(swahili_word);
        if (dict_pos > -1)  plural_flag = 'y';
        else
        {                     /*  and a few U/MA words  */
          swahili_word[0] = 'u';
          for (i = 1; i < 15; ++i)
            swahili_word[i] = swahili_hold[i+1];
          swahili_word[15] = '\x00';   /* mark end of word */
          dict_pos = trsnoun(swahili_word);
          if (dict_pos > -1)  plural_flag = 'y';
        }
      }
    }
  }
  else if (swahili_hold[0] == 'n' && swahili_hold[1] == 'd')
  {
    swahili_word[0] = 'u';    /*  UL/ND  */
    swahili_word[1] = 'l';
    dict_pos = trsnoun(swahili_word);
    if (dict_pos > -1)  plural_flag = 'y';
  }
  else if (swahili_hold[0] == 'n' && swahili_hold[1] == 'y')
  {
    if (swahili_hold[2] == 'u')    /*  U/NYU  */
    {
      for (i = 0; i < 14; ++i)
      swahili_word[i] = swahili_hold[i+2];   /*word w/o ny-*/
      swahili_word[14] = '\x00';   /* mark end of word */
      dict_pos = trsnoun(swahili_word);
      if (dict_pos > -1)  plural_flag = 'y';
    }
    else
    {
      swahili_word[0] = 'w';    /*  W/NY  */
      for (i = 1; i < 15; ++i)
        swahili_word[i] = swahili_hold[i+1];   /*remove 'y'*/
      swahili_word[15] = '\x00';   /* mark end of word */
      dict_pos = trsnoun(swahili_word);
      if (dict_pos > -1)  plural_flag = 'y';
    }
  }
  else if (swahili_hold[0] == 'm' && swahili_hold[1] == 'b')
  {
    swahili_word[0] = 'u';    /*  UB/MB  */
    dict_pos = trsnoun(swahili_word);
    if (dict_pos > -1)  plural_flag = 'y';
  }
  if (plural_flag == 'n')  /* some other words just drop the 'u' */
  {
    swahili_word[0] = 'u';    /*  -/U  */
    for (i = 1; i < 16; ++i)
      swahili_word[i] = swahili_hold[i-1];   /*'u' plus original*/
    dict_pos = trsnoun(swahili_word);
    if (dict_pos > -1)  plural_flag = 'y';
  }
  if ((plural_flag == 'y') &&    /* is this really the plural? */
    (strncmp(swahili_hold,dict_noun.plural_prefix,4) != 0))
      plural_flag = 'n';
  if (plural_flag == 'y')
  {
    english_hold = dict_noun.eng_word;
    eng_index = dict_noun.plural_index;
  /*  if irregular plural entry exists, pick it up  */
    if (eng_index > -1)
    {
      eng_index = irrplu(eng_index);
      english_word = eng_irreg_plural;
    }
    else if (eng_index == -1)    /* -1 means add 's' to word */
    {
      i = 0;
      while (english_hold[i] != '\x00')
      {
        english_word[i] = english_hold[i];
        ++i;
      }
      english_word[i] = 's';
      while (i < 18)    /* clear rest of english_word */
      {
        ++i;
        english_word[i] = '\x00';
      }
    }
    else if (eng_index == -2)    /* -2 means eng plural same */
      english_word = english_hold;
    cout << "\n" << english_word;
    cout << "\n" << "noun - plural" << "\n";
  }
  return plural_flag;
}
//*  Adjective form check routine
char adjform (void)
{
  char adjform_found = 'n';
  char modify_adj[14];
//*  Check for an adjective or other word modifying a noun.
//*  For each type, check is word exists and changes forms.
  if ((swahili_hold[0] == 'k') &&
     ((swahili_hold[1] == 'u') ||
     ((swahili_hold[1] == 'w') &&
      (swahili_hold[2] == 'a' || swahili_hold[2] == 'e'  ||
       swahili_hold[2] == 'o'))))
  {
    for (i = 0; i < 14; ++i)
      swahili_word[i] = swahili_hold[i+2];   /*word w/o ku-, kw- */
    swahili_word[14] = '\x00';   /* mark end of word */
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type > 0))
    {
      cout << "\n" << dict_adjetc.eng_word;
      cout << "\n" << part_speech << " - modifying ";
      cout << "gerund or indicating to/from place" << "\n";
      adjform_found = 'y';
    }
  }
  else  if ((swahili_hold[0] == 'k' || swahili_hold[0] == 'v')  &&
            (swahili_hold[1] == 'i'))
  {
    for (i = 0; i < 14; ++i)
      swahili_word[i] = swahili_hold[i+2];   /*word w/o ki-, vi-*/
    swahili_word[14] = '\x00';   /* mark end of word */
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type > 0))
    {
      cout << "\n" << dict_adjetc.eng_word;
      if (swahili_hold[0] == 'v')
        modify_adj = "plural KI";
      else modify_adj = "KI";
      cout << "\n" << part_speech << " - modifying ";
      cout << modify_adj << " class noun" << "\n";
      adjform_found = 'y';
    }
  }
  else if (((swahili_hold[0] == 'c' && swahili_hold[1] == 'h')  ||
           (swahili_hold[0] == 'v' && swahili_hold[1] == 'y')) &&
           (swahili_hold[2] == 'a' || swahili_hold[2] == 'e'  ||
            swahili_hold[2] == 'o'))
  {
    for (i = 0; i < 14; ++i)
      swahili_word[i] = swahili_hold[i+2];   /*word w/o ch-, vy- */
    swahili_word[14] = '\x00';   /* mark end of word */
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type > 0))
    {
      cout << "\n" << dict_adjetc.eng_word;
      if (swahili_hold[0] == 'v')
        modify_adj = "plural KI";
      else modify_adj = "KI";
      cout << "\n" << part_speech << " - modifying ";
      cout << modify_adj << " class noun" << "\n";
      adjform_found = 'y';
    }
  }
  else if ((swahili_hold[0] == 'j') &&
          ((swahili_hold[1] == 'e') || (swahili_hold[1] == 'i')))
  {
    for (i = 0; i < 15; ++i)
      swahili_word[i] = swahili_hold[i+1];   /*word w/o j-*/
    swahili_word[15] = '\x00';   /* mark end of word */
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
    {
      cout << "\n" << dict_adjetc.eng_word;
      cout << "\n" << part_speech << " - modifying ";
      cout << "MA class noun" << "\n";
      adjform_found = 'y';
    }
  }
  else if ((swahili_hold[0] == 'm') && (swahili_hold[1] != 'e'))
  {
    i = 0;
    if ((swahili_hold[1] == 'i') || (swahili_hold[1] == 'a') ||
        (swahili_hold[1] == 'w'))
    {
      for (; i < 14; ++i)
        swahili_word[i] = swahili_hold[i+2];   /*word w/o ma/mi/mw-*/
      swahili_word[14] = '\x00';   /* mark end of word */
    }
    else
    {
      for (; i < 15; ++i)
        swahili_word[i] = swahili_hold[i+1];   /*word w/o m-*/
      swahili_word[15] = '\x00';   /* mark end of word */
      if (swahili_hold[1] == 'b' && swahili_hold[2] == 'i' &&
          swahili_hold[3] == 'l')
        swahili_word[0] = 'w';    /*  wili/mbili  */
    }
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
    {
      cout << "\n" << dict_adjetc.eng_word;
      if (swahili_hold[1] == 'a')
        modify_adj = "plural MA";
      else if (swahili_hold[1] == 'i')
        modify_adj = "plural MI";
      else if (swahili_hold[1] == 'b')
        modify_adj = "M or N";
      else modify_adj = "M or U";
      cout << "\n" << part_speech << " - modifying ";
      cout << modify_adj << " class noun" << "\n";
      adjform_found = 'y';
    }
  }
  else if (swahili_hold[0] == 'w' && swahili_hold[1] == 'a')
  {
    for (i = 0; i < 14; ++i)
      swahili_word[i] = swahili_hold[i+2];   /*word w/o wa-*/
    swahili_word[14] = '\x00';   /* mark end of word */
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
    {
      cout << "\n" << dict_adjetc.eng_word;
      modify_adj = "plural WA";
      cout << "\n" << part_speech << " - modifying ";
      cout << modify_adj << " class noun" << "\n";
      adjform_found = 'y';
    }
  }
  else if ((swahili_hold[0] == 'm' || swahili_hold[0] == 'w' ||
            swahili_hold[0] == 'p') && (swahili_hold[1] == 'e'))
  {        /*  ma-, wa-, pa- + e- or i- makes me-, we-, pe-  */
    if (swahili_hold[0] == 'm')
      modify_adj = "plural MA";
    else if (swahili_hold[0] == 'w')
      modify_adj = "plural WA";
    else modify_adj = "PA";
    for (i = 0; i < 15; ++i)
      swahili_word[i] = swahili_hold[i+1];   /*word w/o m-, w-*/
    swahili_word[15] = '\x00';   /* mark end of word */
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
    {
      cout << "\n" << dict_adjetc.eng_word;
      cout << "\n" << part_speech << " - modifying ";
      cout << modify_adj << " class noun" << "\n";
      adjform_found = 'y';
    }
    else     /*  maybe the adjective starts with i-  */
    {
      swahili_word[0] = 'i';
      dict_pos = trsadj(swahili_word);
      if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
      {
        cout << "\n" << dict_adjetc.eng_word;
        cout << "\n" << part_speech << " - modifying ";
        cout << modify_adj << " class noun" << "\n";
        adjform_found = 'y';
      }
    }
  }
  else if ((swahili_hold[0] == 'n')  &&
    (swahili_hold[1] == 'd' || swahili_hold[1] == 'g' ||
     swahili_hold[1] == 'j' || swahili_hold[1] == 'z' ))
  {
    for (i = 0; i < 15; ++i)
      swahili_word[i] = swahili_hold[i+1];   /*word w/o n-*/
    swahili_word[15] = '\x00';   /* mark end of word */
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
    {
      cout << "\n" << dict_adjetc.eng_word;
      modify_adj = "N or plural U";
      cout << "\n" << part_speech << " - modifying ";
      cout << modify_adj << " class noun" << "\n";
      adjform_found = 'y';
    }
    else if (swahili_hold[1] == 'd')
    {
      swahili_word[0] = 'r';  /* 'n' plus 'r' becomes 'nd' */
      dict_pos = trsadj(swahili_word);
      if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
      {
        cout << "\n" << dict_adjetc.eng_word;
        modify_adj = "N or plural U";
        cout << "\n" << part_speech << " - modifying ";
        cout << modify_adj << " class noun" << "\n";
        adjform_found = 'y';
      }
    }
    else if ((swahili_hold[1] == 'j') && (swahili_hold[2] == 'e'))
    {
      for (i = 0; i < 14; ++i)
        swahili_word[i] = swahili_hold[i+2];   /*word w/o nj-*/
      swahili_word[14] = '\x00';   /* mark end of word */
      dict_pos = trsadj(swahili_word);
      if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
      {
        cout << "\n" << dict_adjetc.eng_word;
        modify_adj = "N";
        cout << "\n" << part_speech << " - modifying ";
        cout << modify_adj << " class noun" << "\n";
        adjform_found = 'y';
      }
    }
  }
  else if ((swahili_hold[0] == 'n') && (swahili_hold[1] == 'y'))
  {
    for (i = 0; i < 14; ++i)
      swahili_word[i] = swahili_hold[i+2];   /*word w/o ny-*/
    swahili_word[14] = '\x00';   /* mark end of word */
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
    {
      cout << "\n" << dict_adjetc.eng_word;
      modify_adj = "N";
      cout << "\n" << part_speech << " - modifying ";
      cout << modify_adj << " class noun" << "\n";
      adjform_found = 'y';
    }
  }
  else if (swahili_hold[0] == 'p' && swahili_hold[1] == 'a')
  {
    for (i = 0; i < 14; ++i)
      swahili_word[i] = swahili_hold[i+2];   /*word w/o pa-*/
    swahili_word[14] = '\x00';   /* mark end of word */
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type == 1))
    {
      cout << "\n" << dict_adjetc.eng_word;
      modify_adj = "PA";
      cout << "\n" << part_speech << " - modifying ";
      cout << modify_adj << " class noun" << "\n";
      adjform_found = 'y';
    }
  }
  return adjform_found;
}
//*  Special word form check routine
char specform (void)
{
  char specform_found = 'n';
  char subject_phrase[21];
  int letter_count = 0;
  if ((swahili_hold[0] == 'y') && (swahili_hold[1] == 'u'))
  {
    subject_phrase = "M/WA";
    letter_count = 2;
  }
//* distinguish between wa-le and w-ake (et al)
  else if ((swahili_hold[0] == 'w') && (swahili_hold[1] == 'a') &&
      (swahili_hold[2] == 'l'))
  {
    subject_phrase = "plural WA";
    letter_count = 2;
  }
  else if (swahili_hold[0] == 'u')
  {
    subject_phrase = "M/MI or U";
    letter_count = 1;
  }
  else if (swahili_hold[0] == 'i')
  {
    subject_phrase = "N or plural MI";
    letter_count = 1;
  }
  else if ((swahili_hold[0] == 'z') && (swahili_hold[1] == 'i'))
  {
    subject_phrase = "plural N or plural U";
    letter_count = 2;
  }
  else if ((swahili_hold[0] == 'l') && (swahili_hold[1] == 'i') &&
           (swahili_hold[2] != 'a'))    /* lia is not valid */
  {
    subject_phrase = "MA";
    letter_count = 2;
  }
//* distinguish between ya-le and y-ake (et al)
  else if ((swahili_hold[0] == 'y') && (swahili_hold[1] == 'a') &&
           (swahili_hold[2] == 'l'))
  {
    subject_phrase = "plural MA";
    letter_count = 2;
  }
//* distinguish between pa-le and p-ake (et al)
  else if ((swahili_hold[0] == 'p') && (swahili_hold[1] == 'a') &&
           (swahili_hold[2] == 'l'))
  {
    subject_phrase = "PA";
    letter_count = 2;
  }
  else if (swahili_hold[0] == 'w')
  {
    subject_phrase = "M or plural WA";
    letter_count = 1;
  }
  else if (swahili_hold[0] == 'y')
  {
    subject_phrase = "N or plural MI or MA";
    letter_count = 1;
  }
  else if (swahili_hold[0] == 'z')
  {
    subject_phrase = "plural N or U";
    letter_count = 1;
  }
  else if (swahili_hold[0] == 'l')
  {
    subject_phrase = "MA";
    letter_count = 1;
  }
  else if (swahili_hold[0] == 'p')
  {
    subject_phrase = "PA";
    letter_count = 1;
  }
  if (letter_count > 0)    /*  found special prefix  */
  {                           /*  strip subject prefix  */
    for (i = 0; i < (16 - letter_count); ++i)
      swahili_word[i] = swahili_hold[i+letter_count];
    swahili_word[i] = '\x00';
    dict_pos = trsadj(swahili_word);
    if ((dict_pos > -1) && (dict_adjetc.forms_type == 2))
    {
      cout << "\n" << dict_adjetc.eng_word;
      cout << "\n" << part_speech << " - modifying ";
      cout << subject_phrase << " class noun" << "\n";
      specform_found = 'y';
    }
  }
  return specform_found;
}
//*  Check for this and that (referred to)
char thisthat (void)
{
  char thisthat_found = 'n';
  char subject_phrase[21];
  if ((swahili_hold[0] != 'h') || (strlen(swahili_hold) > 5))
    return 'n';
  if (strlen(swahili_hold) == 3)
  {
    if ((swahili_hold[1] == 'a') && (swahili_hold[2] == 'o'))
    {
      thisthat_found = 'b';
      subject_phrase = "plural WA";
    }
    else if ((swahili_hold[1] == 'i') && (swahili_hold[2] == 'i'))
    {
      thisthat_found = 'a';
      subject_phrase = "N or plural MI";
    }
    else if (swahili_hold[1] == 'u')
    {
      subject_phrase = "M/MI or U";
      if (swahili_hold[2] == 'o')
        thisthat_found = 'b';
      else if (swahili_hold[2] == 'u')
        thisthat_found = 'a';
    }
  }
  else if (strlen(swahili_hold) == 4)
  {
    if ((swahili_hold[1] == 'a') && (swahili_hold[2] == 'w') &&
        (swahili_hold[3] == 'a'))
    {
      subject_phrase = "plural WA";
      thisthat_found = 'a';
    }
    else if ((swahili_hold[1] == 'a') && (swahili_hold[2] == 'y'))
    {
      subject_phrase = "plural MA";
      if (swahili_hold[3] == 'a')
        thisthat_found = 'a';
      else if (swahili_hold[3] == 'o')
        thisthat_found = 'b';
    }
    else if (swahili_hold[1] == 'i')
    {
      if ((swahili_hold[2] == 'k') && (swahili_hold[3] == 'i'))
      {
        subject_phrase = "KI";
        thisthat_found = 'a';
      }
      else if (swahili_hold[2] == 'l')
      {
        subject_phrase = "MA";
        if (swahili_hold[3] == 'i')
          thisthat_found = 'a';
        else if (swahili_hold[3] == 'o')
          thisthat_found = 'b';
      }
      else if ((swahili_hold[2] == 'v') && (swahili_hold[3] == 'i'))
      {
        subject_phrase = "plural VI";
        thisthat_found = 'a';
      }
      else if ((swahili_hold[2] == 'y') && (swahili_hold[3] == 'o'))
      {
        subject_phrase = "N or plural MI";
        thisthat_found = 'b';
      }
      else if (swahili_hold[2] == 'z')
      {
        subject_phrase = "plural N or plural U";
        if (swahili_hold[3] == 'i')
          thisthat_found = 'a';
        else if (swahili_hold[3] == 'o')
          thisthat_found = 'b';
      }
    }
    else if ((swahili_hold[1] == 'u') && (swahili_hold[2] == 'y'))
    {
      subject_phrase = "M/WA";
      if (swahili_hold[3] == 'o')
        thisthat_found = 'b';
      else if (swahili_hold[3] == 'u')
        thisthat_found = 'a';
    }
  }
  else if (swahili_hold[4] == 'o')
  {
    if ((swahili_hold[2] == 'c') && (swahili_hold[3] == 'h'))
    {
      thisthat_found = 'b';
      subject_phrase = "KI";
    }
    else if ((swahili_hold[2] == 'v') && (swahili_hold[3] == 'y'))
    {
      thisthat_found = 'b';
      subject_phrase = "plural VI";
    }
  }
  if ((thisthat_found == 'a') || (thisthat_found == 'b'))
  {
    if (thisthat_found == 'a')  cout << "\n" << "this, these";
      else  cout << "\n" << "that, those (referred to)";
    cout << "\n" << "adjective - modifying ";
    cout << subject_phrase << " class noun" << "\n";
    thisthat_found = 'y';
  }
  return thisthat_found;
}
//*  Verb form check routine
char verbform (void)
{
  char verbform_found = 'n';
  char subject_phrase[46];
  char subject_word[11];
  int word_length;
  int letter_count, tense_letter_count;
  int subject_index = 0;
  int simple_subject_index = 0;
  int tense_index = 0;
  word_length = strlen(swahili_hold);
//*  first infinitive and gerund
  if ((swahili_hold[0] == 'k') && ((swahili_hold[1] == 'u') ||
     ((swahili_hold[1] == 'w') && (swahili_hold[2] == 'e' ||
       swahili_hold[2] == 'i' ||
       (swahili_hold[2] == 'a' && swahili_hold[3] == 'n' &&
        swahili_hold[4] == 'z')))))
  {
    for (i = 0; i < 14; ++i)
      swahili_word[i] = swahili_hold[i+2];   /*word w/o ku-*/
    swahili_word[14] = '\x00';   /* mark end of word */
    dict_pos = trsverb(swahili_word);
    if (dict_pos > -1)
    {
      cout << "\n" << "to " << dict_verb.eng_word;
      cout << ", " << dict_verb.eng_ing_form;
      cout << "\n" << "verb - infinitive, gerund" << "\n";
      return 'y';
    }
    else return 'n';
  }
//*  next plural imperative - ends in "-ni"
  if ((swahili_hold[word_length-2] == 'n') &&
      (swahili_hold[word_length-1] == 'i'))
  {
    for (i = 0; i < (word_length-2); ++i)
      swahili_word[i] = swahili_hold[i];   /*word w/o -ni*/
    swahili_word[word_length-2] = '\x00';   /* mark end of word */
    dict_pos = trsverb(swahili_word);
    if (dict_pos > -1)  verbform_found = 'y';
    else if (swahili_hold[word_length-3] == 'e')
    {                                      /* -a becomes -eni */
      swahili_word[word_length-3] = 'a';
      dict_pos = trsverb(swahili_word);
      if (dict_pos > -1)  verbform_found = 'y';
    }
    if (verbform_found == 'y')
    {
//*  Make first letter of English upper-case, for imperative
      if (dict_verb.eng_word[0] > '\x60' &&
        dict_verb.eng_word[0] < '\x7B')
          dict_verb.eng_word[0] = dict_verb.eng_word[0] - '\x20';
      cout << "\n" << dict_verb.eng_word << "!";
      cout << "\n" << "verb - plural imperative" << "\n";
      return 'y';
    }
  }
//*  next customary action tense
  if ((swahili_hold[0] == 'h') && (swahili_hold[1] == 'u'))
  {
    for (i = 0; i < 14; ++i)
      swahili_word[i] = swahili_hold[i+2];   /*word w/o hu-*/
    swahili_word[14] = '\x00';   /* mark end of word */
    dict_pos = trsverb(swahili_word);
    if (dict_pos > -1)
    {
      if (dict_verb.grammar_flag == 'b')    /* verb takes 'to be' */
        cout << "\n" << "is, are " << dict_verb.eng_s_form;
      else
      {
        cout << "\n" << dict_verb.eng_word << ", ";
        cout << dict_verb.eng_s_form;
      }
      cout << "\n" << "verb - customary action tense" << "\n";
      verbform_found = 'y';
    }
    else verbform_found = 'n';
  }
//*  look for subject prefix - simple present and other tenses
  if ((swahili_hold[0] == 'n') && (swahili_hold[1] == 'a'))
  {
    subject_phrase = "first-person singular";
    subject_word = "I";
    letter_count = 2;
    simple_subject_index = 1;
  }
  else if (swahili_hold[0] == 'a')
  {
    subject_phrase = "third-person singular";
    subject_word = "he, she";
    letter_count = 1;
    simple_subject_index = subject_index = 3;
  }
  else if ((swahili_hold[0] == 't') && (swahili_hold[1] == 'w')
        && (swahili_hold[2] == 'a'))
  {
    subject_phrase = "first-person plural";
    subject_word = "we";
    letter_count = 3;
    simple_subject_index = 4;
  }
  else if ((swahili_hold[0] == 'm') && (swahili_hold[1] == 'w')
        && (swahili_hold[2] == 'a'))
  {
    subject_phrase = "second-person plural";
    subject_word = "you";
    letter_count = 3;
    simple_subject_index = 5;
  }
  else if (swahili_hold[0] == 'm')
  {
    subject_phrase = "second-person plural";
    subject_word = "you";
    letter_count = 1;
    subject_index = 5;
  }
  else if ((swahili_hold[0] == 'w') && (swahili_hold[1] == 'a'))
  {
    subject_phrase = "second-person singular OR third-person plural";
    subject_word = "you, they";
    letter_count = 2;
    simple_subject_index = 7;
  }
  else if ((swahili_hold[0] == 'n') && (swahili_hold[1] == 'i'))
  {
    subject_phrase = "first-person singular";
    subject_word = "I";
    letter_count = 2;
    subject_index = 1;
  }
  else if (swahili_hold[0] == 'u')
  {
    subject_phrase = "M/MI class noun";
    subject_word = "(it)";
    letter_count = 1;
    simple_subject_index = 8;
  }
  else if ((swahili_hold[0] == 't') && (swahili_hold[1] == 'u'))
  {
    subject_phrase = "first-person plural";
    subject_word = "we";
    letter_count = 2;
    subject_index = 4;
  }
  else if ((swahili_hold[0] == 'k') && (swahili_hold[1] == 'i'))
  {
    subject_phrase = "KI-class noun";
    subject_word = "(it)";
    letter_count = 2;
    simple_subject_index = subject_index = 3;
  }
  else if ((swahili_hold[0] == 'v') && (swahili_hold[1] == 'i'))
  {
    subject_phrase = "plural KI-class noun";
    subject_word = "(they)";
    letter_count = 2;
    simple_subject_index = subject_index = 6;
  }
  else if ((swahili_hold[0] == 'l') && (swahili_hold[1] == 'i'))
  {
    subject_phrase = "MA-class noun";
    subject_word = "(it)";
    letter_count = 2;
    simple_subject_index = subject_index = 3;
  }
  else if ((swahili_hold[0] == 'p') && (swahili_hold[1] == 'a'))
  {
    subject_phrase = "PA-class noun";
    subject_word = "(it)";
    letter_count = 2;
    simple_subject_index = subject_index = 3;
  }
  else if ((swahili_hold[0] == 'y') && (swahili_hold[1] == 'a'))
  {
    subject_phrase = "plural MA-class noun";
    subject_word = "(they)";
    letter_count = 2;
    simple_subject_index = subject_index = 6;
  }
  else if (swahili_hold[0] == 'i')
  {
    subject_phrase = "N-class or plural MI-class noun";
    subject_word = "(it/they)";
    letter_count = 1;
    simple_subject_index = subject_index = 9;
  }
  else if ((swahili_hold[0] == 'z') && (swahili_hold[1] == 'i'))
  {
    subject_phrase = "plural N-class or plural U-class noun";
    subject_word = "(they)";
    letter_count = 2;
    simple_subject_index = subject_index = 6;
  }
  if (simple_subject_index != 0)    /*  found simple present prefix  */
  {                           /*  strip subject and tense  */
    tense_index = 1;
    for (i = 0; i < (16 - letter_count); ++i)
      swahili_word[i] = swahili_hold[i+letter_count];
    swahili_word[i] = '\x00';
    dict_pos = trsverb(swahili_word);
    if (dict_pos > -1)
    {
      if (dict_verb.grammar_flag == 'b')    /* verb takes 'to be' */
      {
        if (simple_subject_index == 1)
          cout << "\n" << "I am " << dict_verb.eng_s_form;
        else if ((simple_subject_index == 3) ||
                 (simple_subject_index == 8))
          cout << "\n" << subject_word << " is "
            << dict_verb.eng_s_form;
        else if (simple_subject_index == 9)
          cout << "\n" << subject_word << " is/are "
            << dict_verb.eng_s_form;
        else  cout << "\n" << subject_word << " are "
                << dict_verb.eng_s_form;
      }
      else if ((simple_subject_index == 3) ||
               (simple_subject_index == 8))
        cout << "\n" << subject_word << " " << dict_verb.eng_s_form;
      else if (simple_subject_index == 9)
        cout << "\n" << subject_word << " " << dict_verb.eng_s_form
          << "/" << dict_verb.eng_word;
      else  cout << "\n" << subject_word << " " << dict_verb.eng_word;
      cout << "\n" << "verb - simple present tense";
      cout << "\nsubject: " << subject_phrase << "\n";
      return 'y';
    }
    else                 /*  reset tense index and meaning of  */
    {                        /*  certain prefixes  */
      tense_index = 0;
      if (simple_subject_index == 7)
      {
        subject_phrase = "third-person plural";
        subject_word = "they";
        subject_index = 6;
      }
      else if (simple_subject_index == 8)
      {
        subject_phrase = "MI-class noun or second-person singular";
        subject_word = "(it)/you";
        subject_index = 9;
      }
    }
  }
  if ((subject_index > 0) && (tense_index == 0))    /* found subject */
  {                                                 /* but not tense */
    tense_letter_count = 2;    /* for most tenses */
    if ((swahili_hold[letter_count] == 'l') &&
        (swahili_hold[letter_count+1] == 'i'))
      tense_index = 2;    /* past */
    else if ((swahili_hold[letter_count] == 't') &&
             (swahili_hold[letter_count+1] == 'a'))
      tense_index = 3;    /* future */
    else if ((swahili_hold[letter_count] == 'n') &&
             (swahili_hold[letter_count+1] == 'a'))
      tense_index = 4;    /* present */
    else if ((swahili_hold[letter_count] == 'm') &&
             (swahili_hold[letter_count+1] == 'e'))
      tense_index = 5;    /* perfect */
    else if ((swahili_hold[letter_count] == 'k') &&
             (swahili_hold[letter_count+1] == 'a'))
      tense_index = 6;    /* consecutive */
    else if ((swahili_hold[letter_count] == 'k') &&
             (swahili_hold[letter_count+1] == 'i'))
      tense_index = 7;    /* simultaneous */
    else if ((swahili_hold[letter_count] == 'n') &&
             (swahili_hold[letter_count+1] == 'g'))
    {
      if (swahili_hold[letter_count+2] == 'e')
      {
        tense_index = 8;
        tense_letter_count = 3;
      }
      else if ((swahili_hold[letter_count+2] == 'a') &&
               (swahili_hold[letter_count+3] == 'l') &&
               (swahili_hold[letter_count+4] == 'i'))
      {
        tense_index = 9;
        tense_letter_count = 5;
      }
    }
  }
  if (tense_index > 0)    /*  found subject and tense  */
  {                           /*  strip subject and tense  */
    letter_count = letter_count + tense_letter_count;
    for (i = 0; i < (16 - letter_count); ++i)
      swahili_word[i] = swahili_hold[i+letter_count];
    swahili_word[i] = '\x00';
    dict_pos = trsverb(swahili_word);
    if (dict_pos == -1)  /* if no match, maybe ku- or kw- before */
    {                    /*   certain verbs                      */
      if (((swahili_word[0] == 'k') && (swahili_word[1] == 'u') &&
          (swahili_word[3] == 'a') && (swahili_word[4] == '\x00')) ||
          ((swahili_word[0] == 'k') && (swahili_word[1] == 'w') &&
          (swahili_word[2] == 'e' || swahili_word[2] == 'i')))
      {
        for (i = 0; i < 16; ++i)
          swahili_word[i] = swahili_word[i+2];
        dict_pos = trsverb(swahili_word);
      }
    }
    if (dict_pos > -1)
    {
      if (tense_index == 2)
      {
        if (dict_verb.grammar_flag == 'b')    /* verb takes 'to be' */
        {
          if ((subject_index == 1) | (subject_index == 3))
            cout << "\n" << subject_word << " was "
              << dict_verb.eng_past;
          else if (subject_index == 9)
            cout << "\n" << subject_word << " was/were "
              << dict_verb.eng_past;
          else  cout << "\n" << subject_word << " were "
                  << dict_verb.eng_past;
        }
        else cout << "\n" << subject_word << " " << dict_verb.eng_past;
        cout << "\n" << "verb - past tense";
      }
      else if (tense_index == 3)
      {
        cout << "\n" << subject_word << " will " << dict_verb.eng_word;
        cout << "\n" << "verb - future tense";
      }
      else if (tense_index == 4)
      {
        if (subject_index == 1)
          cout << "\n" << "I am " << dict_verb.eng_ing_form;
        else if (subject_index == 3)
          cout << "\n" << subject_word << " is "
            << dict_verb.eng_ing_form;
        else if (subject_index == 9)
          cout << "\n" << subject_word << " is/are "
            << dict_verb.eng_ing_form;
        else  cout << "\n" << subject_word << " are "
                << dict_verb.eng_ing_form;
        cout << "\n" << "verb - present tense";
      }
      else if (tense_index == 5)
      {
      if (subject_index == 3)
        cout << "\n" << subject_word << " has "
          << dict_verb.eng_participle;
      else if (subject_index == 9)
        cout << "\n" << subject_word << " has/have "
          << dict_verb.eng_participle;
      else  cout << "\n" << subject_word << " have "
              << dict_verb.eng_participle;
        cout << "\n" << "verb - perfect tense";
      }
      else if (tense_index == 6)
      {
        if (dict_verb.grammar_flag == 'b')    /* verb takes 'to be' */
        {
          if ((subject_index == 1) | (subject_index == 3))
            cout << "\nand " << subject_word << " was "
              << dict_verb.eng_past;
          else if (subject_index == 9)
            cout << "\nand" << subject_word << " was/were "
              << dict_verb.eng_past;
          else  cout << "\nand " << subject_word << " were "
                  << dict_verb.eng_past;
        }
        else cout << "\nand " << subject_word << " "
               << dict_verb.eng_past;
        cout << "\n" << "verb - consecutive tense";
      }
      else if (tense_index == 7)
      {
        cout << "\nif, when " << subject_word;
        if (dict_verb.grammar_flag == 'b')    /* verb takes 'to be' */
        {
          if (subject_index == 1)  cout << " am ";
          else if (subject_index == 3)  cout << " is ";
          else if (subject_index == 9)  cout << " is/are ";
            else  cout << " are ";
          cout  << dict_verb.eng_s_form;
        }
        else if (simple_subject_index == 3)
          cout << " " << dict_verb.eng_s_form;
        else if (simple_subject_index == 9)
          cout << " " << dict_verb.eng_s_form << "/"
            << dict_verb.eng_word;
          else  cout << " " << dict_verb.eng_word;
        cout << "\n" << "verb - simultaneous tense";
      }
      else if (tense_index == 8)
      {
        cout << "\n" << subject_word << " would "
          << dict_verb.eng_word;
        cout << "\n" << "verb - conditional tense";
      }
      else if (tense_index == 9)
      {
        cout << "\n" << subject_word << " would have "
          << dict_verb.eng_participle;
        cout << "\n" << "verb - past conditional tense";
      }
      cout << "\nsubject: " << subject_phrase << "\n";
      return 'y';
    }
  }
//*  now look for subject prefix - negative tenses
  subject_index = 0;    /* reset indexes */
  tense_index = 0;
  if ((swahili_hold[0] == 's') && (swahili_hold[1] == 'i'))
  {
    subject_phrase = "first-person singular";
    subject_word = "I";
    letter_count = 2;
    subject_index = 1;
  }
  else if ((swahili_hold[0] == 'h') && (swahili_hold[1] == 'u'))
  {
    subject_phrase = "second-person singular";
    subject_word = "you";
    letter_count = 2;
    subject_index = 2;
  }
  else if ((swahili_hold[0] == 'h') && (swahili_hold[1] == 'a'))
  {
    if ((swahili_hold[2] == 't') && (swahili_hold[3] == 'u'))
    {
      subject_phrase = "first-person plural";
      subject_word = "we";
      letter_count = 4;
      subject_index = 4;
    }
    else if (swahili_hold[2] == 'm')
    {
      subject_phrase = "second-person plural";
      subject_word = "you";
      letter_count = 3;
      subject_index = 5;
    }
    else if ((swahili_hold[2] == 'w') && (swahili_hold[3] == 'a'))
    {
      subject_phrase = "third-person plural";
      subject_word = "they";
      letter_count = 4;
      subject_index = 6;
    }
    else if ((swahili_hold[2] == 'k') && (swahili_hold[3] == 'i'))
    {
      subject_phrase = "KI-class noun";
      subject_word = "(it)";
      letter_count = 4;
      subject_index = 3;
    }
    else if ((swahili_hold[2] == 'v') && (swahili_hold[3] == 'i'))
    {
      subject_phrase = "plural KI-class noun";
      subject_word = "(they)";
      letter_count = 4;
      subject_index = 6;
    }
    else if (swahili_hold[2] == 'u')
    {
      subject_phrase = "M/MI class noun";
      subject_word = "(it)";
      letter_count = 3;
      subject_index = 3;
    }
    else if ((swahili_hold[2] == 'l') && (swahili_hold[3] == 'i'))
    {
      subject_phrase = "MA-class noun";
      subject_word = "(it)";
      letter_count = 4;
      subject_index = 3;
    }
    else if ((swahili_hold[2] == 'y') && (swahili_hold[3] == 'a'))
    {
      subject_phrase = "plural MA-class noun";
      subject_word = "(they)";
      letter_count = 4;
      subject_index = 6;
    }
    else if (swahili_hold[2] == 'i')
    {
      subject_phrase = "N-class or plural MI-class noun";
      subject_word = "(it, they)";
      letter_count = 3;
      subject_index = 6;
    }
    else if ((swahili_hold[2] == 'z') && (swahili_hold[3] == 'i'))
    {
      subject_phrase = "plural N-class or plural U-class noun";
      subject_word = "(they)";
      letter_count = 4;
      subject_index = 6;
    }
    else    /* ha- by itself */
    {
      subject_phrase = "third-person singular";
      subject_word = "he, she";
      letter_count = 2;
      subject_index = 3;
    }
  }
  if (subject_index > 0)    /*  found negative subject  */
  {
    if (swahili_hold[word_length-1] == 'i')
      tense_index = 1;    /* present */
    else if ((swahili_hold[letter_count] == 'k') &&
        (swahili_hold[letter_count+1] == 'u'))
      tense_index = 2;    /* past */
    else if ((swahili_hold[letter_count] == 't') &&
        (swahili_hold[letter_count+1] == 'a'))
      tense_index = 3;    /* future */
    else if ((swahili_hold[letter_count] == 'j') &&
        (swahili_hold[letter_count+1] == 'a'))
      tense_index = 4;    /* perfect */
    else if (swahili_hold[word_length-1] != 'a')
      tense_index = 5;    /* might still be present */
  }
  if (tense_index > 0)    /*  found negative tense  */
  {
    if ((tense_index != 1) & (tense_index != 5))
      letter_count = letter_count + 2;
    for (i = 0; i < (16 - letter_count); ++i)
      swahili_word[i] = swahili_hold[i+letter_count];
    swahili_word[i] = '\x00';
      /* if neg present ends with 'i', root verb ends with 'a' */
    if (tense_index == 1)
      swahili_word[word_length-letter_count-1] = 'a';
    dict_pos = trsverb(swahili_word);
    if (dict_pos > -1)
    {
      if ((tense_index == 1) | (tense_index == 5))
      {
        if (dict_verb.grammar_flag == 'b')    /* verb takes 'to be' */
        {
          if (subject_index == 1)
            cout << "\n" << "I am not " << dict_verb.eng_s_form;
          else if (subject_index == 3)
            cout << "\n" << subject_word << " is not "
              << dict_verb.eng_s_form;
          else  cout << "\n" << subject_word << " are not "
                  << dict_verb.eng_s_form;
        }
        else if (subject_index == 3)
          cout << "\n" << subject_word << " does not "
            << dict_verb.eng_word;
        else  cout << "\n" << subject_word << " do not "
          << dict_verb.eng_word;
        cout << "\n" << "verb - negative present tense";
      }
      else if (tense_index == 2)
      {
        if (dict_verb.grammar_flag == 'b')    /* verb takes 'to be' */
        {
          if ((subject_index == 1) | (subject_index == 3))
            cout << "\n" << subject_word << " was not "
              << dict_verb.eng_past;
          else  cout << "\n" << subject_word << " were not "
                  << dict_verb.eng_past;
        }
        else cout << "\n" << subject_word << " did not "
                  << dict_verb.eng_word;
        cout << "\n" << "verb - negative past tense";
      }
      else if (tense_index == 3)
      {
        cout << "\n" << subject_word << " will not "
          << dict_verb.eng_word;
        cout << "\n" << "verb - negative future tense";
      }
      else if (tense_index == 4)
      {
        if (subject_index == 3)
              cout << "\n" << subject_word << " has not yet "
                << dict_verb.eng_participle;
        else  cout << "\n" << subject_word << " have not yet "
                << dict_verb.eng_participle;
        cout << "\n" << "verb - negative perfect tense";
      }
      cout << "\nsubject: " << subject_phrase << "\n";
      return 'y';
    }
  }
  return verbform_found;
}
//* Kiswahili-English dictionary lookup routine - nouns
int trsnoun (char word_in[])
{
  int low_entry = 0;
  int high_entry = control_data.noun_dict_entries - 1;
  int mid_entry;
  int long mid_entry_seek;
//* binary search
  while (low_entry <= high_entry)
  {
    mid_entry = (low_entry + high_entry) / 2;
    mid_entry_seek = mid_entry * NOUNS_ENTRY_SIZE;
    fseek (noun_dict, mid_entry_seek, 0);
    fread (&dict_noun, NOUNS_ENTRY_SIZE, 1, noun_dict);
    if (strcmp(dict_noun.swa_word,word_in) < 0)
      low_entry = mid_entry + 1;
    else if (strcmp(dict_noun.swa_word,word_in) > 0)
      high_entry = mid_entry - 1;
    else return (mid_entry);
  }
  return -1;
}
//* Kiswahili-English dictionary lookup routine - verbs
int trsverb (char word_in[])
{
  int low_entry = 0;
  int high_entry = control_data.verb_dict_entries - 1;
  int mid_entry;
  int long mid_entry_seek;
//* binary search
  while (low_entry <= high_entry)
  {
    mid_entry = (low_entry + high_entry) / 2;
    mid_entry_seek = mid_entry * VERBS_ENTRY_SIZE;
    fseek (verb_dict, mid_entry_seek, 0);
    fread (&dict_verb, VERBS_ENTRY_SIZE, 1, verb_dict);
    if (strcmp(dict_verb.swa_word,word_in) < 0)
      low_entry = mid_entry + 1;
    else if (strcmp(dict_verb.swa_word,word_in) > 0)
      high_entry = mid_entry - 1;
    else return (mid_entry);
  }
  return -1;
}
//* Kiswahili-English dictionary lookup routine - adjectives etc.
int trsadj (char word_in[])
{
  int low_entry = 0;
  int high_entry = control_data.adjetc_dict_entries - 1;
  int mid_entry;
  int long mid_entry_seek;
//* binary search
  while (low_entry <= high_entry)
  {
    mid_entry = (low_entry + high_entry) / 2;
    mid_entry_seek = mid_entry * ADJETC_ENTRY_SIZE;
    fseek (adjetc_dict, mid_entry_seek, 0);
    fread (&dict_adjetc, ADJETC_ENTRY_SIZE, 1, adjetc_dict);
    if (strcmp(dict_adjetc.swa_word,word_in) < 0)
      low_entry = mid_entry + 1;
    else if (strcmp(dict_adjetc.swa_word,word_in) > 0)
      high_entry = mid_entry - 1;
    else    /* found word - convert speech_part, then return */
    {
    if (dict_adjetc.speech_part == 'j')
      part_speech = "adjective";
    else if (dict_adjetc.speech_part == 'c')
      part_speech = "conjunction";
    else if (dict_adjetc.speech_part == 'p')
      part_speech = "pronoun";
    else if (dict_adjetc.speech_part == 'a')
      part_speech = "adverb";
    else if (dict_adjetc.speech_part == 'r')
      part_speech = "preposition";
    else if (dict_adjetc.speech_part == 'v')
      part_speech = "irregular verb form";
    else if (dict_adjetc.speech_part == 'i')
      part_speech = "interjection";
    else if (dict_adjetc.speech_part == 'x')  /* not applicable */
      part_speech = "x";
    else part_speech = "x";
    return (mid_entry);
    }
  }
  return -1;
}
//* English irregular plural lookup routine
int irrplu (int index_in)
{
  int long entry_seek;
  entry_seek = index_in * IRRPLU_ENTRY_SIZE;
  fseek (irrplu_dict, entry_seek, 0);
  fread (eng_irreg_plural, IRRPLU_ENTRY_SIZE, 1, irrplu_dict);
  return index_in;
}
